@@ -3,7 +3,7 @@
 #include "frameobject.h"
 #include "pythread.h"
 
-#define VERSION "0.7"
+#define VERSION "0.8"
 
 #if PY_MAJOR_VERSION >= 3
 # define PYTHON3
@@ -180,7 +180,6 @@ trace_timer_check(void)
 static void
 trace_log_alloc(void *ptr, trace_alloc_t *trace)
 {
-    g_hash_table_insert(trace_allocs, ptr, trace);
     trace_config.enabled = 0;
     trace->filename = trace_get_filename(&trace->lineno);
     trace_config.enabled = 1;
@@ -196,6 +195,8 @@ trace_log_alloc(void *ptr, trace_alloc_t *trace)
         trace->filename = "???";
     }
 
+    g_hash_table_insert(trace_allocs, ptr, trace);
+
     trace_update_stats(1, trace);
 
     trace_timer_check();
@@ -249,12 +250,14 @@ trace_realloc(trace_realloc_t func, void *ptr1, size_t size)
             return func(ptr1, size);
         }
     }
+    else
+        trace1 = NULL;
 
     ptr2 = func(ptr1, size);
     if (ptr2 == NULL)
         return NULL;
 
-    if (ptr1 != NULL)
+    if (trace1 != NULL)
         trace_log_dealloc(ptr1, trace1);
 
     trace2 = trace_alloc_trace();
@@ -285,6 +288,49 @@ trace_free(trace_free_t func, void *ptr)
         trace_log_dealloc(ptr, trace);
 }
 
+#ifdef WITH_FREE_LIST
+static void
+trace_free_list_alloc(PyObject *op)
+{
+    void *ptr;
+    PyTypeObject *type;
+    size_t size;
+    trace_alloc_t *trace;
+
+    if (!trace_config.enabled)
+        return;
+
+    trace = trace_alloc_trace();
+    if (trace == NULL)
+        return;
+
+    type = Py_TYPE(op);
+    if (PyType_IS_GC(type))
+        ptr = (void *)((char *)op - sizeof(PyGC_Head));
+    else
+        ptr = (void *)op;
+    size = type->tp_basicsize;
+
+    trace->size = size;
+    trace_log_alloc(ptr, trace);
+}
+
+static void
+trace_free_list_free(PyObject *op)
+{
+    trace_alloc_t *trace;
+    void *ptr;
+
+    if (!trace_config.enabled)
+        return;
+
+    ptr = (void *)op;
+    trace = g_hash_table_lookup(trace_allocs, ptr);
+    if (trace != NULL)
+        trace_log_dealloc(ptr, trace);
+}
+#endif
+
 static void *
 trace_mem_malloc(size_t size)
 {
@@ -458,6 +504,12 @@ trace_register_allocators(void)
                          trace_object_free) < 0)
         return -1;
 
+#ifdef WITH_FREE_LIST
+    if (_PyFreeList_SetAllocators(trace_free_list_alloc,
+                                  trace_free_list_free) < 0)
+        return -1;
+#endif
+
     return 0;
 }
 
@@ -647,7 +699,7 @@ trace_lineno_as_obj(int lineno)
 }
 
 PyDoc_STRVAR(trace_get_stats_doc,
-    "get_stats() -> dict\n"
+    "_get_stats() -> dict\n"
     "\n"
     "Get allocation statistics per Python file as a dict:\n"
     "{filename (str): {lineno (int) -> (size (int), count (int))}}");
@@ -774,45 +826,101 @@ done:
     return get_stats.file_dict;
 }
 
-PyDoc_STRVAR(trace_get_source_doc,
-    "get_source(obj) -> (filename: str, lineno: int)\n"
+PyDoc_STRVAR(trace_get_object_trace_doc,
+    "_get_object_trace(obj) -> (size: int, filename: str, lineno: int)\n"
     "\n"
-    "Get the source of an object.\n"
-    "Return (filename, lineno) if the source is known,\n"
+    "Get the memory allocation trace of an object.\n"
+    "Return (size, filename, lineno) if the source is known,\n"
     "None otherwise.");
 
 static PyObject*
-py_trace_get_source(PyObject *self, PyObject *obj)
+py_trace_get_object_trace(PyObject *self, PyObject *obj)
 {
-    trace_alloc_t *trace;
+    PyTypeObject *type;
     void *ptr;
-    PyObject *filename, *lineno;
+    trace_alloc_t *trace;
+    PyObject *size = NULL, *filename = NULL, *lineno = NULL;
 
-    ptr = (void *)obj;
+    type = Py_TYPE(obj);
+    if (PyType_IS_GC(type))
+        ptr = (void *)((char *)obj - sizeof(PyGC_Head));
+    else
+        ptr = (void *)obj;
     trace = g_hash_table_lookup(trace_allocs, ptr);
     if (trace == NULL) {
         Py_INCREF(Py_None);
         return Py_None;
     }
 
+    size = PYINT_FROM_SSIZE_T(trace->size);
+    if (size == NULL)
+        goto error;
+
     filename = trace_decode_filename(trace->filename);
     if (filename == NULL)
-        return NULL;
+        goto error;
 
     lineno = trace_lineno_as_obj(trace->lineno);
-    if (lineno == NULL) {
-        Py_DECREF(filename);
-        return NULL;
-    }
-    return Py_BuildValue("(NN)", filename, lineno);
+    if (lineno == NULL)
+        goto error;
+
+    return Py_BuildValue("(NNN)", size, filename, lineno);
+
+error:
+    Py_XDECREF(size);
+    Py_XDECREF(filename);
+    Py_XDECREF(lineno);
+    return NULL;
 }
 
+static int
+trace_atexit_register(PyObject *module)
+{
+    PyObject *disable = NULL, *atexit = NULL, *func = NULL;
+    PyObject *result;
+    int ret = -1;
+
+    disable = PyObject_GetAttrString(module, "disable");
+    if (disable == NULL)
+        goto done;
+
+    atexit = PyImport_ImportModule("atexit");
+    if (atexit == NULL) {
+        if (!PyErr_Warn(PyExc_ImportWarning,
+                       "atexit module is missing: "
+                       "cannot automatically disable tracemalloc at exit"))
+        {
+            PyErr_Clear();
+            return 0;
+        }
+        goto done;
+    }
+
+    func = PyObject_GetAttrString(atexit, "register");
+    if (func == NULL)
+        goto done;
+
+    result = PyObject_CallFunction(func, "O", disable);
+    if (result == NULL)
+        goto done;
+    Py_DECREF(result);
+
+    ret = 0;
+
+done:
+    Py_XDECREF(disable);
+    Py_XDECREF(func);
+    Py_XDECREF(atexit);
+    return ret;
+}
+
+
 
 static PyMethodDef trace_methods[] = {
     {"enable", (PyCFunction)py_trace_enable, METH_NOARGS, trace_enable_doc},
     {"disable", (PyCFunction)py_trace_disable, METH_NOARGS, trace_disable_doc},
-    {"get_source", (PyCFunction)py_trace_get_source, METH_O, trace_get_source_doc},
-    {"get_stats", (PyCFunction)py_trace_get_stats, METH_NOARGS, trace_get_stats_doc},
+    {"_get_object_trace", (PyCFunction)py_trace_get_object_trace, METH_O, trace_get_object_trace_doc},
+    {"_get_stats", (PyCFunction)py_trace_get_stats, METH_NOARGS, trace_get_stats_doc},
     {"start_timer", py_trace_start_timer, METH_VARARGS, trace_start_timer_doc},
     {"stop_timer", (PyCFunction)py_trace_stop_timer, METH_NOARGS, trace_stop_timer_doc},
     {NULL,              NULL}           /* sentinel */
@@ -844,35 +952,38 @@ init_tracemalloc(void)
 {
     PyObject *m, *version;
 
-    if (trace_init() < 0) {
-#ifdef PYTHON3
-        return NULL;
-#else
-        return;
-#endif
-    }
+    if (trace_init() < 0)
+        goto error;
 
 #ifdef PYTHON3
     m = PyModule_Create(&sandbox_module);
-    if (m == NULL)
-        return NULL;
 #else
     m = Py_InitModule3("_tracemalloc", trace_methods, trace_doc);
-    if (m == NULL)
-        return;
 #endif
+    if (m == NULL)
+        goto error;
 
 #ifdef PYTHON3
     version = PyUnicode_FromString(VERSION);
-    if (version == NULL)
-        return NULL;
 #else
     version = PyString_FromString(VERSION);
 #endif
+    if (version == NULL)
+        goto error;
     PyModule_AddObject(m, "__version__", version);
 
+    if (trace_atexit_register(m) < 0)
+        goto error;
+
 #ifdef PYTHON3
     return m;
 #endif
+
+error:
+#ifdef PYTHON3
+    return NULL;
+#else
+    return;
+#endif
 }
 
